from jamboree.handlers.default import DataHandler

class EconomicData(DataHandler):
    """
        # Economic Data
        
        A way to browse economic data. Is an extension of DataHandler and includes basic searches.
    """
    pass