import unittest


"""
    The things tested in this document:
    ---
    1. Storing a lot of information at a single level
        * This is a save_many command on a dataframe
    2. Querying information and resampling it for processing it.
    3. Initializing a certain set of data to process at different layers.
        * This entails getting small data and resampling it prior to querying
        * Query resampled data
"""


