from abc import ABC
from typing import List

class SearchProcessor(ABC):
    """ All of the common search queries will go here """
    def search(self):
        pass


    def search_specific(self):
        pass

    def search_all(self):
        """ Search all of the tags"""
        pass